"""replace key with value in sample table and display unicode as the corresponding Chinese

Revision ID: 9d5da133bbe4
Revises: e76c2ca5562e
Create Date: 2023-03-07 10:41:33.590250

"""
import json

from alembic import op
from alembic import context
from sqlalchemy import update
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.automap import automap_base

def get_tool_label_dict(task_config: dict) -> dict:
    """get the key value of labels in a given task_id and task_config"""

    label_dict = {"无标签": "noneAttribute"}
    # obtain the labels in current task
    # get the general labels
    for normal_label in task_config.get("attribute", []):
        if normal_label.get("key", ""):
            label_dict[normal_label.get("key")] = normal_label.get("value")
    # get the labels in configuration defined by user
    for task_tool in task_config.get("tools", []):
        if "config" not in task_tool.keys():
            continue
        labels = task_tool.get("config").get("attributeList", [])
        for label in labels:
            if label.get("key", ""):
                label_dict[label.get("key")] = label.get("value")
    return label_dict


def replace_key_with_value(sample_data: dict, label_dict: dict) -> dict:
    """replace the key with value in task_sample table to modify the error for the history version"""

    annotated_result = sample_data.get("result")
    annotated_result = json.loads(annotated_result)
    for sample_tool, sample_tool_results in annotated_result.items():
        if sample_tool.endswith("Tool"):
            for sample_tool_result in sample_tool_results.get("result", []):
                tool_label = sample_tool_result.get("attribute", "")
                if tool_label in label_dict:
                    sample_tool_result["attribute"] = label_dict[tool_label]
    return annotated_result

# revision identifiers, used by Alembic.
revision = "9d5da133bbe4"
down_revision = "e76c2ca5562e"
branch_labels = None
depends_on = None

# produce a set of mappings from this MetaData
Base = automap_base()


def upgrade() -> None:
    """Replace key with value in sample table"""

    # get the current connection
    bind = op.get_bind()
    # Reflect ORM models from the database
    Base.prepare(autoload_with=bind, reflect=True)
    task_sample = Base.classes.task_sample
    # make a session
    Session = sessionmaker(bind=bind)
    session = Session()

    # replace the key with value in a transaction
    with context.begin_transaction():
        task_items = session.execute(
            'SELECT id, config FROM task WHERE config IS NOT NULL AND config != ""'
        )
        for task_item in task_items:
            task_id = task_item[0]
            task_config = json.loads(task_item[1])
            label_dict = get_tool_label_dict(task_config)
            # replace key with value in the sample of current task
            # get the sample data items of the task id
            sample_items = session.execute(
                f"SELECT id, data, annotated_count FROM task_sample WHERE task_id={task_id}"
            )
            # if task have several annotated sample, continute to replace key with value
            for sample_item in sample_items:
                sample_id = sample_item[0]
                sample_data_item = json.loads(sample_item[1])
                # replace key with value and display unicode as the corresponding Chinesedisplays unicode as the corresponding Chinese
                if sample_item[2] > 0:
                    sample_annotated_result = replace_key_with_value(
                        sample_data_item, label_dict
                    )
                    sample_data_item["result"] = json.dumps(
                        sample_annotated_result, ensure_ascii=False
                    )
                # display unicode as the corresponding Chinese
                else:
                    pass
                sample_annotated_item_str = json.dumps(
                    sample_data_item, ensure_ascii=False
                )
                op.execute(
                    update(task_sample)
                    .where(task_sample.id == sample_id)
                    .values({task_sample.data: sample_annotated_item_str})
                )


def downgrade() -> None:
    pass



"""add_sample_name_and_data_to_task_pre_annotation

1. Add sample_name column to task_pre_annotation table.
2. Read jsonl file content and store it in data column of task_pre_annotation table.

Revision ID: eb9c5b98168b
Revises: bc8fcb35b66b
Create Date: 2024-11-13 14:08:09.374271

"""
import json
from typing import List
from alembic import op, context
import sqlalchemy as sa
from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy.ext.automap import automap_base
from sqlalchemy import select
from labelu.internal.common.config import settings

from labelu.internal.domain.models.pre_annotation import TaskPreAnnotation
from labelu.internal.adapter.persistence import crud_attachment

Base = automap_base()


# revision identifiers, used by Alembic.
revision = 'eb9c5b98168b'
down_revision = 'bc8fcb35b66b'
branch_labels = None
depends_on = None

def index_exists(bind, table_name, index_name):
    inspector = sa.inspect(bind)
    indexes = inspector.get_indexes(table_name)
    for index in indexes:
        if index['name'] == index_name:
            return True
    return False

def read_jsonl_file(db: Session, file_id: int) -> List[dict]:
    attachment = crud_attachment.get(db, file_id)
    if attachment is None:
        return []

    attachment_path = attachment.path
    file_full_path = settings.MEDIA_ROOT.joinpath(attachment_path.lstrip("/"))
    
    # check if the file exists
    if not file_full_path.exists() or not attachment.filename.endswith('.jsonl'):
        return []

    
    with open(file_full_path, "r", encoding="utf-8") as f:
        data = f.readlines()

    parsed_data = [json.loads(line) for line in data]
    return parsed_data

def upgrade() -> None:
    bind = op.get_bind()
    Base.prepare(autoload_with=bind, reflect=True)
    # make a session
    SessionMade = sessionmaker(bind=bind)
    session = SessionMade()
    
    try:
        with context.begin_transaction():
            # add column sample_name, check if the column exists
            if not hasattr(Base.classes.task_pre_annotation, "sample_name"):
                op.add_column(
                    "task_pre_annotation",
                    sa.Column("sample_name", sa.String(255), index=True, comment="One of the sample names of the task"),
                )
            
            # create index, check if the index exists
            if not index_exists(bind, "task_pre_annotation", "idx_pre_annotation_sample_name"):
                op.create_index("idx_pre_annotation_sample_name", "task_pre_annotation", ["sample_name"])
                
            # create task_pre_annotation
            exist_task_pre_annotations = session.execute(
                select([Base.classes.task_pre_annotation])
            ).scalars().all()
            
            for task_pre_annotation in exist_task_pre_annotations:
                file_id = task_pre_annotation.file_id
                jsonl_contents = read_jsonl_file(session, file_id)
                
                # create new task_pre_annotation
                for jsonl_content in jsonl_contents:
                    sample_name = jsonl_content.get("sample_name")
                    query = select(Base.classes.task_attachment).where(
                        Base.classes.task_attachment.task_id == task_pre_annotation.task_id,
                        # filename include sample_name, full name is xxxxxxxxx-sample_name.png, shot name is sample_name.png
                        Base.classes.task_attachment.filename.contains(sample_name),
                    )
                    sample_file = session.execute(query).scalars().first()
                    new_task_pre_annotation = TaskPreAnnotation(
                        task_id=task_pre_annotation.task_id,
                        # full file name
                        sample_name=sample_file.filename if sample_file else None,
                        file_id=file_id,
                        created_by=task_pre_annotation.created_by,
                        updated_by=task_pre_annotation.updated_by,
                        data=json.dumps(jsonl_content),
                    )
                    session.add(new_task_pre_annotation)
            
            # remove existing task_pre_annotation
            for task_pre_annotation in exist_task_pre_annotations:
                session.delete(task_pre_annotation)
                
            # commit 
            session.commit()
            
    except Exception as e:
        session.rollback()
        raise e
    
    finally:
        session.close()


def downgrade() -> None:
    bind = op.get_bind()
    Base.prepare(autoload_with=bind, reflect=True)
    SessionMade = sessionmaker(bind=bind)
    session = SessionMade()
    
    try:
        with context.begin_transaction():
            # remove new task_pre_annotation
            new_task_pre_annotations = session.query(Base.classes.task_pre_annotation).filter(
                Base.classes.task_pre_annotation.sample_name.isnot(None)
            ).all()
            
            for task_pre_annotation in new_task_pre_annotations:
                session.delete(task_pre_annotation)
            
            # restore old task_pre_annotation
            old_task_pre_annotations = session.query(TaskPreAnnotation).filter(
                TaskPreAnnotation.sample_name.isnot(None)
            ).all()
            
            for task_pre_annotation in old_task_pre_annotations:
                restored_task_pre_annotation = Base.classes.task_pre_annotation(
                    task_id=task_pre_annotation.task_id,
                    file_id=task_pre_annotation.file_id,
                    created_by=task_pre_annotation.created_by,
                    updated_by=task_pre_annotation.updated_by,
                    data=task_pre_annotation.data,
                )
                session.add(restored_task_pre_annotation)
            
            # drop index
            if index_exists(bind, "task_pre_annotation", "idx_pre_annotation_sample_name"):
                op.drop_index("idx_pre_annotation_sample_name", table_name="task_pre_annotation")
            
            # drop column
            if hasattr(Base.classes.task_pre_annotation, "sample_name"):
                op.drop_column("task_pre_annotation", "sample_name")
            
            # commit
            session.commit()
    
    except Exception as e:
        session.rollback()
        raise e
    
    finally:
        session.close()


import json
import os
from PIL import Image
from typing import List
from tfrecord import example_pb2
from labelu.internal.common.config import settings

class TF_record_converter:
    def __init__(self):
        self.label_map = {}
        
    def _get_label_id(self, label: str):
        if label not in self.label_map:
            self.label_map[label] = len(self.label_map) + 1
            
        return self.label_map[label]
        
    def create_tf_examples(self, sample_results: List[dict], config: dict):
        examples = []
        label_text_dict = {}
        
        for tool in config.get("tools", []):
            label_text_dict[tool.get("tool")] = { attr.get("value"): attr.get("key") for attr in tool.get("config", {}).get("attributes", [])}
        
        common_attributes = { attr.get("value"): attr.get("key") for attr in config.get("attributes", [])}
        
        def get_label(_tool: str, _input_label: str):
            _label = label_text_dict.get(_tool, {}).get(_input_label, "")
            
            if not _label:
                _label = common_attributes.get(_input_label, "")
                
            return _label
        
        classes_text = []
        classes = []
        
        for sample in sample_results:
            data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            
            # skip invalid data
            annotated_result = json.loads(data.get("result"))
            if sample.get("state") == "SKIPPED" or not annotated_result:
                continue
            
            image_width = annotated_result.get("width", 0)
            image_height = annotated_result.get("height", 0)
            rotate = annotated_result.get("rotate", 0)
            
            _, file_extension = os.path.splitext(file.get("filename", ""))
            
            file_full_path = settings.MEDIA_ROOT.joinpath(file.get("path").lstrip("/"))
            
            with Image.open(file_full_path) as img:
                if rotate:
                    img = img.rotate(rotate)

                image_width, image_height = img.size
                encoded_image_data = img.tobytes()
            
            xmins = []
            xmaxs = []
            ymins = []
            ymaxs = []
            line_x = []
            line_y = []
            segmentation = []
            keypoints_x = []
            keypoints_y = []
            keypoints_visibility = []
            cuboid_front = []
            cuboid_back = []
            tags = []
            texts = []
            
            base_info = {
                "image/width": self._int64_feature(image_width),
                "image/height": self._int64_feature(image_height),
                "image/filename": self._bytes_feature(file['filename'].encode('utf8')),
                "image/source_id": self._bytes_feature(file['filename'].encode('utf8')),
                "image/encoded": self._bytes_feature(encoded_image_data),
                "image/format": self._bytes_feature(file_extension.encode('utf8')),
            }

            for tool in annotated_result.copy().keys():
                tool_results = annotated_result.pop(tool)
                if tool == "rectTool":
                    for tool_result in tool_results.get("result", []):
                        label_text = get_label(tool, tool_result.get("label", ""))
                        classes_text.append(label_text.encode('utf8'))
                        classes.append(self._get_label_id(label_text))
                        xmins.append(tool_result.get("x", 0) / image_width)
                        xmaxs.append((tool_result.get("x", 0) + tool_result.get("width", 0)) / image_width)
                        ymins.append(tool_result.get("y", 0) / image_height)
                        ymaxs.append((tool_result.get("y", 0) + tool_result.get("height", 0)) / image_height)

                    base_info.update(
                        {
                            "image/object/bbox/xmin": self._float_list_feature(xmins),
                            "image/object/bbox/xmax": self._float_list_feature(xmaxs),
                            "image/object/bbox/ymin": self._float_list_feature(ymins),
                            "image/object/bbox/ymax": self._float_list_feature(ymaxs),
                        }
                    )
                if tool == "lineTool":
                    for tool_result in tool_results.get("result", []):
                        points = tool_result.get("points", [])
                        
                        label_text = get_label(tool, tool_result.get("label", ""))
                        classes.append(label_text.encode('utf8'))
                        classes.append(self._get_label_id(label_text))
                        
                        for point in points:
                            line_x.append(point.get("x", 0) / image_width)
                            line_y.append(point.get("y", 0) / image_height)

                    base_info.update(
                        {
                            "image/object/line/x": self._float_list_feature(line_x),
                            "image/object/line/y": self._float_list_feature(line_y),
                        }
                    )
                    
                if tool == "polygonTool":
                    for tool_result in tool_results.get("result", []):
                        points = tool_result.get("points", [])
                        segmentation = []
                        for point in points:
                            segmentation.append(point.get("x", 0) / image_width)
                            segmentation.append(point.get("y", 0) / image_height)
                        label = tool_result.get("label", "")
                        classes_text.append(label.encode('utf8'))
                        classes.append(self._get_label_id(label))
                        
                    base_info.update(
                        {
                            "image/object/segmentation": self._float_list_feature(segmentation),
                        }
                    )
                
                if tool == "pointTool":
                    for tool_result in tool_results.get("result", []):
                        points = tool_result.get("points", [])
                        for point in points:
                            keypoints_x.append(point.get("x", 0) / image_width)
                            keypoints_y.append(point.get("y", 0) / image_height)
                            # TODO: visibility
                            keypoints_visibility.append(2)
                            
                        label = tool_result.get("label", "")
                        classes_text.append(label.encode('utf8'))
                        classes.append(self._get_label_id(label))
                        
                    base_info.update(
                        {
                            "image/object/keypoints/x": self._float_list_feature(keypoints_x),
                            "image/object/keypoints/y": self._float_list_feature(keypoints_y),
                            "image/object/keypoints/visibility": self._int64_feature(keypoints_visibility),
                        }
                    )
                        
                if tool == "cuboidTool":
                    for tool_result in tool_results.get("result", []):
                        front = tool_result.get("front", {})
                        back = tool_result.get("back", {})
                        for key in ["tl", "tr", "br", "bl"]:
                            cuboid_front.append(front.get(key, {}).get("x", 0) / image_width)
                            cuboid_front.append(front.get(key, {}).get("y", 0) / image_height)
                            cuboid_back.append(back.get(key, {}).get("x", 0) / image_width)
                            cuboid_back.append(back.get(key, {}).get("y", 0) / image_height)
                        label = tool_result.get("label", "")
                        classes_text.append(label.encode('utf8'))
                        classes.append(self._get_label_id(label))
                        
                    base_info.update(
                        {
                            "image/object/cuboid/front": self._float_list_feature(cuboid_front),
                            "image/object/cuboid/back": self._float_list_feature(cuboid_back),
                        }
                    )
                
                if tool == "tagTool":
                    for tool_result in tool_results.get("result", []):
                        for _, tag in tool_result.get("value", {}).items():
                            for value in tag:
                                tags.append(value.encode('utf8'))
                    
                    base_info.update(
                        {
                            "image/object/tag": self._bytes_feature(tags),
                        }
                    )

                if tool == "textTool":
                    for tool_result in tool_results.get("result", []):
                        for _, text in tool_result.get("value", {}).items():
                            texts.append(text.encode('utf8'))
                            
                    base_info.update(
                        {
                            "image/object/text": self._bytes_feature(texts),
                        }
                    )
                    
            base_info.update(
                {
                    "image/object/class/text": self._bytes_feature(classes_text),
                    "image/object/class/label": self._int64_feature(classes),
                }
            )
            
            examples.append(example_pb2.Example(features=example_pb2.Features(feature=base_info)))
        
        return examples            
        
    
    def _bytes_feature(self, value: bytes | List[bytes]):
        return example_pb2.Feature(bytes_list=example_pb2.BytesList(value=[value] if isinstance(value, bytes) else value))

    def _float_list_feature(self, value: List[float]):
        return example_pb2.Feature(float_list=example_pb2.FloatList(value=value))

    def _int64_feature(self, value: int | List[int]):
        return example_pb2.Feature(int64_list=example_pb2.Int64List(value=[value] if isinstance(value, int) else value))



import base64
import csv
import json
import os
from zipfile import ZipFile
from PIL import Image, ImageDraw
from enum import Enum

import xml.etree.ElementTree as ET
from typing import List
from fastapi import status
from loguru import logger

from labelu.internal.common.error_code import ErrorCode, LabelUException

from .xml_converter import XML_converter
from .tf_record_converter import TF_record_converter
from .color import colors
from .config import settings


class Format(str, Enum):
    JSON = "JSON"
    COCO = "COCO"
    MASK = "MASK"
    YOLO = "YOLO"
    CSV = "CSV"
    XML = "XML"
    VOC = "VOC"
    TF_RECORD = "TF_RECORD"
    LABEL_ME = "LABEL_ME"
    PASCAL_VOC = "PASCAL_VOC"


class Converter:
    def convert(
        self,
        config: dict,
        input_data: List[dict],
        out_data_dir: str,
        out_data_file_name_prefix: str,
        format: str,
    ) -> str:
        if format == Format.JSON.value:
            return self.convert_to_json(
                input_data=input_data,
                out_data_dir=out_data_dir,
                out_data_file_name_prefix=out_data_file_name_prefix,
            )
        elif format == Format.COCO.value:
            return self.convert_to_coco(
                config=config,
                input_data=input_data,
                out_data_dir=out_data_dir,
                out_data_file_name_prefix=out_data_file_name_prefix,
            )
        elif format == Format.MASK.value:
            return self.convert_to_mask(
                input_data=input_data,
                out_data_dir=out_data_dir,
                out_data_file_name_prefix=out_data_file_name_prefix,
            )
        elif format == Format.LABEL_ME.value:
            return self.convert_to_labelme(
                config=config,
                input_data=input_data,
                out_data_file_name_prefix=out_data_file_name_prefix,
                out_data_dir=out_data_dir,
            )
        elif format == Format.YOLO.value:
            return self.convert_to_yolo(
                config=config,
                input_data=input_data,
                out_data_file_name_prefix=out_data_file_name_prefix,
                out_data_dir=out_data_dir,
            )
        elif format == Format.CSV.value:
            return self.convert_to_csv(
                config=config,
                input_data=input_data,
                out_data_file_name_prefix=out_data_file_name_prefix,
                out_data_dir=out_data_dir,
            )
        elif format == Format.XML.value:
            return self.convert_to_xml(
                config=config,
                input_data=input_data,
                out_data_file_name_prefix=out_data_file_name_prefix,
                out_data_dir=out_data_dir,
            )
        elif format == Format.TF_RECORD.value:
            return self.convert_to_tf_record(
                config=config,
                input_data=input_data,
                out_data_file_name_prefix=out_data_file_name_prefix,
                out_data_dir=out_data_dir,
            )
        elif format == Format.PASCAL_VOC.value:
            return self.convert_to_pascal_voc(
                config=config,
                input_data=input_data,
                out_data_file_name_prefix=out_data_file_name_prefix,
                out_data_dir=out_data_dir,
            )
            
    def convert_to_json(
        self,
        input_data: List[dict],
        out_data_dir: str,
        out_data_file_name_prefix: str,
    ) -> str:
        out_data_dir.mkdir(parents=True, exist_ok=True)
        file_full_path = out_data_dir.joinpath("result.json")

        # every file result
        results = []
        for sample in input_data:
            data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            
            # change skipped result is invalid
            annotated_result = json.loads(data.get("result"))
            if annotated_result and sample.get("state") == "SKIPPED":
                annotated_result["valid"] = False

            # change result struct
            if annotated_result:
                annotations = []
                for tool in annotated_result.copy().keys():
                    if tool.endswith("Tool"):
                        tool_results = annotated_result.pop(tool)
                        for tool_result in tool_results.get("result", []):
                            # 视频文件的标注结果已经保存了 label 键的值，不需要再做转换
                            if "label" not in tool_result:
                                tool_result["label"] = tool_result.pop("attribute", "")

                            tool_result.pop("sourceID", None)

                            if tool == "tagTool" or tool == "textTool":
                                tool_result.pop("label")

                            if "attribute" in tool_result:
                                tool_result.pop("attribute")

                        annotations.append(tool_results)

                annotated_result["annotations"] = annotations

            annotated_result_str = json.dumps(annotated_result, ensure_ascii=False)
            results.append(
                {
                    "id": sample.get("id"),
                    "result": annotated_result_str,
                    "folder": settings.MEDIA_ROOT,
                    "url": file.get("url"),
                    "fileName": file.get("filename", ""),
                }
            )

        # Serializing json
        json_object = json.dumps(results, default=str)
        with file_full_path.open("w") as outfile:
            outfile.write(json_object)

        logger.info("Export file path: {}", file_full_path)
        return file_full_path

    def convert_to_coco(
        self,
        config: dict,
        input_data: List[dict],
        out_data_dir: str,
        out_data_file_name_prefix: str,
    ) -> str:

        # result output file
        out_data_dir.mkdir(parents=True, exist_ok=True)
        file_full_path = out_data_dir.joinpath("result.json")

        # result struct
        result = {
            "images": [],
            "annotations": [],
            "categories": [],
        }

        # result catetory
        category_id = 0
        logger.info("get categories")
        for attr in config.get("attributes", []):
            category = {
                "id": category_id,
                "name": attr.get("value", ""),
                "supercategory": "",
            }
            result["categories"].append(category)
            category_id += 1
        tools_category = config.get("tools", [])
        for tool in tools_category:
            for attr in tool.get("config", {}).get("attributes", []):
                category = {
                    "id": category_id,
                    "name": attr.get("value", ""),
                    "supercategory": "",
                }
                result["categories"].append(category)
                category_id += 1

        logger.info("get categories map with id")
        category_name_map_id = {}
        for category in result["categories"]:
            category_name_map_id[category.get("name")] = category.get("id")

        # annotation index
        annotation_id = 0

        # for every annotation media
        for sample in input_data:
            annotation_data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            logger.info("data is: {}", sample)

            # annotation result
            annotation_result = json.loads(annotation_data.get("result", {}))

            # coco image
            image = {
                "id": sample.get("id"),
                "fileName": file.get("filename", ""),
                "width": annotation_result.get("width", 0),
                "height": annotation_result.get("height", 0),
                "valid": False
                if sample.get("state", "") == "SKIPPED"
                else annotation_result.get("valid", True),
                "rotate": annotation_result.get("rotate", 0),
            }
            result["images"].append(image)

            # every image may have multi tools
            tools = []
            if annotation_result.get("polygonTool", {}):
                tools.append(annotation_result.get("polygonTool"))
            if annotation_result.get("rectTool", {}):
                tools.append(annotation_result.get("rectTool"))

            for tool in tools:
                for tool_result in tool.get("result"):

                    bbox = []
                    segmentation = []
                    polygon_area = 0.0
                    # polygon tool
                    if tool.get("toolName") == "polygonTool":
                        x_coordinates = []
                        y_coordinates = []
                        for point in tool_result.get("points", []):
                            segmentation.append(point.get("x"))
                            segmentation.append(point.get("y"))
                            x_coordinates.append(point.get("x"))
                            y_coordinates.append(point.get("y"))

                        bbox = [
                            min(x_coordinates),
                            max(y_coordinates),
                            max(x_coordinates) - min(x_coordinates),
                            max(y_coordinates) - min(y_coordinates),
                        ]
                        polygon_area = _polygonArea(x_coordinates, y_coordinates)
                    elif tool.get("toolName") == "rectTool":
                        # rect
                        x = tool_result.get("x")
                        y = tool_result.get("y")
                        width = tool_result.get("width")
                        height = tool_result.get("height")
                        
                        if x is not None and y is not None and width is not None and height is not None:
                            bbox.extend([x, y, width, height])
                        polygon_area = tool_result.get("width", 0) * tool_result.get(
                            "height", 0
                        )

                    annotation = {
                        "image_id": sample.get("id"),
                        "id": annotation_id,
                        "iscrowd": tool_result.get("iscrowd", 0),
                        "segmentation": segmentation,
                        "area": polygon_area,
                        "bbox": bbox,
                        "category_id": category_name_map_id.get(
                            tool_result.get("label", ""), -1
                        ),
                        "order": tool_result.get("order", 0),
                    }

                    annotation_id += 1
                    result["annotations"].append(annotation)

        # Serializing json
        json_object = json.dumps(result, default=str)
        with file_full_path.open("w") as outfile:
            outfile.write(json_object)
        logger.info("Export file path: {}", file_full_path)
        return file_full_path

    def convert_to_mask(
        self,
        input_data: List[dict],
        out_data_dir: str,
        out_data_file_name_prefix: str,
    ) -> str:

        # result output file
        out_data_dir.mkdir(parents=True, exist_ok=True)

        polygon = []
        export_files = []
        color_list = []
        for sample in input_data:
            file = sample.get("file", {})
            if sample.get("state") != "DONE":
                continue
            annotation_data = json.loads(sample.get("data"))
            logger.info("data is: {}", sample)
            filename = file.get("filename")
            if filename and filename.split("/")[-1]:
                file_relative_path_base_name = filename.split("/")[-1].split(".")[0]
            else:
                file_relative_path_base_name = "result"

            # annotation result
            annotation_result = json.loads(annotation_data.get("result", {}))
            if not annotation_result or not annotation_result.get("polygonTool", {}):
                continue

            # polygon tool
            polygons = []
            polygon_attribute = []
            for tool_result in annotation_result.get("polygonTool", {}).get(
                "result", []
            ):
                polygon = []
                for point in tool_result.get("points", []):
                    polygon.append(point.get("x"))
                    polygon.append(point.get("y"))
                polygons.append(polygon)
                polygon_attribute.append(tool_result.get("label", ""))

            width = annotation_result.get("width")
            height = annotation_result.get("height")

            # generate single change
            file_relative_path_model_l = f"{file_relative_path_base_name}-trainIds.png"
            file_full_path_model_l = out_data_dir.joinpath(file_relative_path_model_l)
            img_model_l = Image.new("L", (width, height), 0)
            for _index, p in enumerate(polygons):
                ImageDraw.Draw(img_model_l).polygon(p, outline=1, fill=_index + 1)
            img_model_l.save(file_full_path_model_l, "PNG")
            export_files.append(file_full_path_model_l)

            # generate RGB
            file_relative_path_model_rgb = (
                f"{file_relative_path_base_name}-segmentation.png"
            )
            file_full_path_model_rgb = out_data_dir.joinpath(
                file_relative_path_model_rgb
            )
            img_model_rgb = Image.new("RGB", (width, height), 0)
            for index, p in enumerate(polygons):
                color = colors[index % 255 + 1]
                ImageDraw.Draw(img_model_rgb).polygon(p, fill=color.get("hexString"))

                rgb = color.get("rgb")
                color_list.append(
                    {
                        "color": f'rgb({rgb.get("r")},{rgb.get("g")},{rgb.get("b")})',
                        "colorList": [rgb.get("r"), rgb.get("g"), rgb.get("b"), 255],
                        "trainIds": index + 1,
                        "attribute": polygon_attribute[index],
                    }
                )
            img_model_rgb.save(file_full_path_model_rgb, "PNG")
            export_files.append(file_full_path_model_rgb)

        # color list
        file_full_path_colors = out_data_dir.joinpath("colors.json")
        json_object = json.dumps(color_list, default=str)
        with file_full_path_colors.open("w") as outfile:
            outfile.write(json_object)
        export_files.append(file_full_path_colors)

        file_relative_path_zip = f"task-{out_data_file_name_prefix}-mask.zip"
        file_full_path_zip = out_data_dir.joinpath(file_relative_path_zip)
        with ZipFile(file_full_path_zip, "w") as zipf:
            for f in export_files:
                zipf.write(str(f), arcname=f.name)
        logger.info("Export file path: {}", file_full_path_zip)
        return file_full_path_zip

    def convert_to_labelme(self, config: dict, input_data: List[dict], out_data_file_name_prefix: str, out_data_dir: str):
        out_data_dir.mkdir(parents=True, exist_ok=True)
        export_files = []
        result = []
        # does not support cuboid / spline
        shape_dict = {
            "polygonTool": "polygon",
            "rectTool": "rectangle",
            "lineTool": "linestrip",
            "pointTool": "point",
        }
        
        label_text_dict = {}
        common_attributes = { attr.get("value"): attr.get("key") for attr in config.get("attributes", [])}
        
        for tool in config.get("tools", []):
            label_text_dict[tool.get("tool")] = { attr.get("value"): attr.get("key") for attr in tool.get("config", {}).get("attributes", [])}
        
        def get_label(_tool: str, _input_label: str):
            _label = label_text_dict.get(_tool, {}).get(_input_label, "")
            
            if not _label:
                _label = common_attributes.get(_input_label, "")
                
            return _label
        
        def convert_points(points: List[dict]):
            return [[point.get("x"), point.get("y")] for point in points]
        
        def image_to_base64(file_path: str):
            file_full_path = settings.MEDIA_ROOT.joinpath(file_path.lstrip("/"))
            with open(file_full_path, "rb") as image_file:
                return base64.b64encode(image_file.read()).decode('utf-8')
        
        for sample in input_data:
            labelme_item = {
                "version": "5.5.0",
                "flags": {},
                "shapes": [],
                "imagePath": "",
                "imageData": "",
                "imageHeight": 0,
                "imageWidth": 0,
            }
            data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            
            # skip invalid data
            annotated_result = json.loads(data.get("result"))
            if sample.get("state") == "SKIPPED":
                continue
            
            labelme_item["imagePath"] = file.get("filename", "")
            labelme_item["imageData"] = image_to_base64(file.get("path"))
            
            if annotated_result:
                labelme_item["imageWidth"] = annotated_result.get("width", 0)
                labelme_item["imageHeight"] = annotated_result.get("height", 0)
                
                for tool in annotated_result.copy().keys():
                    if tool.endswith("Tool") and tool in shape_dict:
                        # polygon
                        if tool == "polygonTool":
                            tool_results = annotated_result.pop(tool)
                            for tool_result in tool_results.get("result", []):
                                attributes = tool_result.get("attributes", {})
                                shape = {
                                    "label": get_label(tool, tool_result.get("label", "")),
                                    "points": convert_points(tool_result.get("points", [])),
                                    "group_id": "",
                                    # Get description from attributes
                                    "description": attributes.get("description", ""),
                                    "shape_type": shape_dict.get(tool, "polygon"),
                                    "flags": {},
                                    "mask": "",
                                }
                                labelme_item["shapes"].append(shape)
                                
                        # rect
                        if tool == "rectTool":
                            tool_results = annotated_result.pop(tool)
                            for tool_result in tool_results.get("result", []):
                                attributes = tool_result.get("attributes", {})
                                x = tool_result.get("x", 0)
                                y = tool_result.get("y", 0)
                                width = tool_result.get("width", 0)
                                height = tool_result.get("height", 0)
                                shape = {
                                    "label": get_label(tool, tool_result.get("label", "")),
                                    "points": [[x, y], [x + width, y + height]],
                                    "group_id": "",
                                    "description": attributes.get("description", ""),
                                    "shape_type": shape_dict.get(tool, "rectangle"),
                                    "flags": {},
                                    "mask": "",
                                }
                                labelme_item["shapes"].append(shape)
                        
                        if tool == "lineTool":
                            tool_results = annotated_result.pop(tool)
                            for tool_result in tool_results.get("result", []):
                                attributes = tool_result.get("attributes", {})
                                shape = {
                                    "label": get_label(tool, tool_result.get("label", "")),
                                    "points": convert_points(tool_result.get("points", [])),
                                    "group_id": "",
                                    "description": attributes.get("description", ""),
                                    "shape_type": shape_dict.get(tool, "linestrip"),
                                    "flags": {},
                                    "mask": "",
                                }
                                labelme_item["shapes"].append(shape)
                                
                        if tool == "pointTool":
                            tool_results = annotated_result.pop(tool)
                            for tool_result in tool_results.get("result", []):
                                attributes = tool_result.get("attributes", {})
                                shape = {
                                    "label": get_label(tool, tool_result.get("label", "")),
                                    "points": [[tool_result.get("x", 0), tool_result.get("y", 0)]],
                                    "group_id": "",
                                    "description": attributes.get("description", ""),
                                    "shape_type": shape_dict.get(tool, "point"),
                                    "flags": {},
                                    "mask": "",
                                }
                                labelme_item["shapes"].append(shape)
            result.append(labelme_item)
            file_basename = os.path.splitext(file.get("filename", ""))[0]
            file_name = out_data_dir.joinpath(f"{file_basename}.json")
            with file_name.open("w") as outfile:
                # 格式化json，两个空格缩进
                json.dump(labelme_item, outfile, indent=2, ensure_ascii=False)
                
            export_files.append(file_name)
        
        
        file_relative_path_zip = f"task-{out_data_file_name_prefix}-label-me.zip"
        file_full_path_zip = out_data_dir.joinpath(file_relative_path_zip)
        with ZipFile(file_full_path_zip, "w") as zipf:
            for f in export_files:
                zipf.write(str(f), arcname=f.name)
        logger.info("Export file path: {}", file_full_path_zip)
        return file_full_path_zip
    
    def convert_to_yolo(self, config: dict, input_data: List[dict], out_data_file_name_prefix: str, out_data_dir: str):
        out_data_dir.mkdir(parents=True, exist_ok=True)
        export_files = []
        classes = []
        
        # make classes
        for tool in config.get("tools", []):
            for attr in tool.get("config", {}).get("attributes", []):
                classes.append(attr.get("value"))

        for attr in config.get("attributes", []):
            classes.append(attr.get("value"))
        
        # make classes.txt
        classes_file = out_data_dir.joinpath("classes.txt")
        with classes_file.open("w") as outfile:
            for c in classes:
                outfile.write(f"{c}\n")
        export_files.append(classes_file)
        
        for sample in input_data:
            data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            
            # skip invalid data
            annotated_result = json.loads(data.get("result"))
            if sample.get("state") == "SKIPPED" or not annotated_result:
                continue
            
            image_path = settings.MEDIA_ROOT.joinpath(file.get("path").lstrip("/"))
            file_basename = os.path.splitext(file.get("filename", ""))[0]
            file_name = out_data_dir.joinpath(f"{file_basename}.txt")
            image_width = annotated_result.get("width", 0)
            image_height = annotated_result.get("height", 0)
            rotate = annotated_result.get("rotate", 0)
            
            with Image.open(image_path) as img:
                if rotate:
                    img = img.rotate(rotate, expand=True)
                image_width, image_height = img.size
            
            with file_name.open("a") as outfile:
                if rotate:
                    outfile.write(f"# rotate: {rotate}\n")
                    
                for tool in annotated_result.copy().keys():
                    if tool == 'rectTool':
                        tool_results = annotated_result.pop(tool)
                        for tool_result in tool_results.get("result", []):
                            x = tool_result.get("x", 0)
                            y = tool_result.get("y", 0)
                            width = tool_result.get("width", 0)
                            height = tool_result.get("height", 0)
                            label = tool_result.get("label", "")
                            x_center = x + width / 2
                            y_center = y + height / 2
                            x_center /= image_width
                            y_center /= image_height
                            width /= image_width
                            height /= image_height
                            
                            outfile.write(f"{classes.index(label)} {x_center} {y_center} {width} {height}\n")
            export_files.append(file_name)
            
            
        file_relative_path_zip = f"task-{out_data_file_name_prefix}-yolo.zip"
        file_full_path_zip = out_data_dir.joinpath(file_relative_path_zip)
        with ZipFile(file_full_path_zip, "w") as zipf:
            for f in export_files:
                zipf.write(str(f), arcname=f.name)
        logger.info("Export file path: {}", file_full_path_zip)
        return file_full_path_zip

    def convert_to_csv(self, config: dict, input_data: List[dict], out_data_file_name_prefix: str, out_data_dir: str):
        out_data_dir.mkdir(parents=True, exist_ok=True)
        export_files = []
        
        label_text_dict = {}
        common_attributes = { attr.get("value"): attr.get("key") for attr in config.get("attributes", [])}
        
        def get_label(_tool: str, _input_label: str):
            _label = label_text_dict.get(_tool, {}).get(_input_label, "")
            
            if not _label:
                _label = common_attributes.get(_input_label, "")
                
            return _label
        
        def get_attributes(attributes: dict):
            result = []
            
            for value in attributes.values():
                result.append(", ".join(value) if isinstance(value, list) else value)
                 
            return ", ".join(result)
        
        def get_points(direction: dict):
            return [
                (
                    direction.get("tl").get("x"),
                    direction.get("tl").get("y"),
                ),
                (
                    direction.get("tr").get("x"),
                    direction.get("tr").get("y"),
                ),
                (
                    direction.get("br").get("x"),
                    direction.get("br").get("y"),
                ),
                (
                    direction.get("bl").get("x"),
                    direction.get("bl").get("y"),
                ),
            ]
                
        
        for tool in config.get("tools", []):
            label_text_dict[tool.get("tool")] = { attr.get("value"): attr.get("key") for attr in tool.get("config", {}).get("attributes", [])}
            
        for sample in input_data:
            data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            # tool_name, label, x, y, width, height etc.
            rows = []
            
            # skip invalid data
            annotated_result = json.loads(data.get("result"))
            if sample.get("state") == "SKIPPED" or not annotated_result:
                continue
            
            for tool in annotated_result.copy().keys():
                if tool == 'rectTool':
                    tool_results = annotated_result.pop(tool)
                    rows.append(["tool_name", "label", "label_text", "x", "y", "width", "height", "attributes", "order"])
                    for tool_result in tool_results.get("result", []):
                        x = tool_result.get("x", 0)
                        y = tool_result.get("y", 0)
                        width = tool_result.get("width", 0)
                        height = tool_result.get("height", 0)
                        label = tool_result.get("label", "")
                        order = tool_result.get("order", 0)
                        label_text = get_label(tool, label)
                        rows.append([tool, label, label_text, x, y, width, height, get_attributes(tool_result.get('attributes', {})), order])
                        
                if tool == 'lineTool':
                    tool_results = annotated_result.pop(tool)
                    rows.append(["tool_name", "label", "label_text", "points", "control_points", "attributes", "order"])
                    for tool_result in tool_results.get("result", []):
                        points = tool_result.get("points", [])
                        control_points = tool_result.get("controlPoints", [])
                        label = tool_result.get("label", "")
                        order = tool_result.get("order", 0)
                        label_text = get_label(tool, label)
                        rows.append([tool, label, label_text, points, control_points, get_attributes(tool_result.get('attributes', {})), order])
                        
                if tool == 'pointTool':
                    tool_results = annotated_result.pop(tool)
                    rows.append(["tool_name", "label", "label_text", "x", "y", "order"])
                    for tool_result in tool_results.get("result", []):
                        x = tool_result.get("x", 0)
                        y = tool_result.get("y", 0)
                        label = tool_result.get("label", "")
                        order = tool_result.get("order", 0)
                        label_text = get_label(tool, label)
                        rows.append([tool, label, label_text, x, y, order])
                        
                if tool == 'polygonTool':
                    tool_results = annotated_result.pop(tool)
                    rows.append(["tool_name", "label", "label_text", "points", "attributes", "order"])
                    for tool_result in tool_results.get("result", []):
                        points = tool_result.get("points", [])
                        control_points = tool_result.get("controlPoints", [])
                        label = tool_result.get("label", "")
                        order = tool_result.get("order", 0)
                        label_text = get_label(tool, label)
                        rows.append([tool, label, label_text, points, control_points, get_attributes(tool_result.get('attributes', {})), order])
                
                if tool == 'cuboidTool':
                    tool_results = annotated_result.pop(tool)
                    rows.append(["tool_name", "label", "label_text", "direction", "front", "back", "attributes", "order"])
                    for tool_result in tool_results.get("result", []):
                        direction = tool_result.get("direction")
                        # [[x,y], ...]
                        front = get_points(tool_result.get("front"))
                        back = get_points(tool_result.get("back"))
                        width = tool_result.get("width", 0)
                        height = tool_result.get("height", 0)
                        label = tool_result.get("label", "")
                        order = tool_result.get("order", 0)
                        label_text = get_label(tool, label)
                        rows.append([tool, label, label_text, direction, front, back, get_attributes(tool_result.get('attributes', {})), order])
                        
            file_basename = os.path.splitext(file.get("filename", ""))[0]
            file_name = out_data_dir.joinpath(f"{file_basename}.csv")
            with file_name.open("w") as outfile:
                writer = csv.writer(outfile)
                writer.writerows(rows)
            export_files.append(file_name)
            
        file_relative_path_zip = f"task-{out_data_file_name_prefix}-csv.zip"
        file_full_path_zip = out_data_dir.joinpath(file_relative_path_zip)
        with ZipFile(file_full_path_zip, "w") as zipf:
            for f in export_files:
                zipf.write(str(f), arcname=f.name)
        logger.info("Export file path: {}", file_full_path_zip)
        return file_full_path_zip


    def convert_to_xml(self, config: dict, input_data: List[dict], out_data_file_name_prefix: str, out_data_dir: str):
        out_data_dir.mkdir(parents=True, exist_ok=True)
        file_full_path = out_data_dir.joinpath("result.xml")
        
        # result struct
        xml_converter = XML_converter()
        root = ET.Element("root")
        
        for sample in input_data:
            data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            sample_item = ET.SubElement(root, "sample")
            
            # skip invalid data
            annotated_result = json.loads(data.get("result"))
            result = ET.SubElement(sample_item, "result")
            if annotated_result and sample.get("state") == "SKIPPED":
                ET.SubElement(result, "valid").text = "False"

            ET.SubElement(sample_item, "id").text = str(sample.get("id"))
            
            ET.SubElement(sample_item, "folder").text = str(settings.MEDIA_ROOT)
            ET.SubElement(sample_item, "path").text = file.get("path")
            ET.SubElement(sample_item, "fileName").text = file.get("filename", "")
            
            ET.SubElement(result, "width").text = str(annotated_result.get("width", 0))
            ET.SubElement(result, "height").text = str(annotated_result.get("height", 0))
            ET.SubElement(result, "rotate").text = str(annotated_result.get("rotate", 0))
            
            # change result struct
            if annotated_result:
                annotations = ET.SubElement(result, "annotations")
                for tool in annotated_result.copy().keys():
                    tool_results = annotated_result.pop(tool)
                    if tool.endswith("Tool"):
                        tool_annotations = xml_converter.convert_tool_results(tool, tool_results)
                
                        for annotation in tool_annotations:
                            annotations.append(annotation)
               
        tree = ET.ElementTree(root)
        with open(file_full_path, 'wb') as f:
            tree.write(f, encoding='utf-8', xml_declaration=True)
        logger.info("Export file path: {}", file_full_path)
        return file_full_path
    
    def convert_to_tf_record(self, config: dict, input_data: List[dict], out_data_file_name_prefix: str, out_data_dir: str):
        out_data_dir.mkdir(parents=True, exist_ok=True)
        export_files = []
        
        # result struct
        tf_record_examples = TF_record_converter().create_tf_examples(input_data, config)
        
        if len(tf_record_examples) == 0:
            raise LabelUException(
                code=ErrorCode.CODE_61000_NO_DATA,
                status_code=status.HTTP_400_BAD_REQUEST,
            )
        
        for sample in input_data:
            file = sample.get("file", {})
            if len(tf_record_examples) == 0:
                continue
            example = tf_record_examples.pop(0)
            file_basename = os.path.splitext(file.get("filename", ""))[0]
            tf_record = f"{file_basename}.tfrecord"
            file_full_path = out_data_dir.joinpath(tf_record)
            
            with file_full_path.open("wb") as outfile:
                outfile.write(example.SerializeToString())
                
            export_files.append(file_full_path)
        
        file_relative_path_zip = f"task-{out_data_file_name_prefix}-tfrecord.zip"
        file_full_path_zip = out_data_dir.joinpath(file_relative_path_zip)
        with ZipFile(file_full_path_zip, "w") as zipf:
            for f in export_files:
                zipf.write(str(f), arcname=f.name)
        logger.info("Export file path: {}", file_full_path_zip)
        return file_full_path_zip
    
    def convert_to_pascal_voc(self, config: dict, input_data: List[dict], out_data_file_name_prefix: str, out_data_dir: str):
        out_data_dir.mkdir(parents=True, exist_ok=True)
        export_files = []
        
        xml_converter = XML_converter()
        
        for sample in input_data:
            data = json.loads(sample.get("data"))
            file = sample.get("file", {})
            
            # skip invalid data
            annotated_result = json.loads(data.get("result"))
            if sample.get("state") == "SKIPPED" or not annotated_result:
                continue
            
            voc_xml = xml_converter.create_pascal_voc_xml(config, file, annotated_result)
            file_basename = os.path.splitext(file.get("filename", ""))[0]
            file_name = out_data_dir.joinpath(f"{file_basename}.xml")
            
            tree = ET.ElementTree(voc_xml)
            tree.write(file_name, encoding="utf-8", xml_declaration=True)
            
            export_files.append(file_name)
    
        file_relative_path_zip = f"task-{out_data_file_name_prefix}-pascal-voc.zip"
        file_full_path_zip = out_data_dir.joinpath(file_relative_path_zip)
        
        with ZipFile(file_full_path_zip, "w") as zipf:
            for f in export_files:
                zipf.write(str(f), arcname=f.name)
        logger.info("Export file path: {}", file_full_path_zip)
        return file_full_path_zip
    
def _polygonArea(X, Y):

    # Initialize area
    area = 0.0
    n = len(X)

    # Calculate value of shoelace formula
    j = n - 1
    for i in range(0, n):
        area += (X[j] + X[i]) * (Y[j] - Y[i])
        j = i  # j is previous vertex to i

    # Return absolute value
    return abs(area) / 2.0

converter = Converter()



from pathlib import Path
import random
from labelu.internal.common.config import settings
import string


def random_lower_string() -> str:
    return "".join(random.choices(string.ascii_lowercase, k=16))


def random_username() -> str:
    return f"{random_lower_string()}@{random_lower_string()}.com"

def empty_task_upload(task_id: int, input_filename: str):
    path_filename = input_filename.split("/")
    #  filename = str(uuid.uuid4())[0:8] + "-" + path_filename[-1]
    filename = path_filename[-1]
    path = "/".join(path_filename[:-1])
    # delete task 1 folder
    attachment_relative_base_dir = Path(settings.UPLOAD_DIR).joinpath(
        str(task_id), path
    )
    attachment_relative_path = str(attachment_relative_base_dir.joinpath(filename))

    attachment_full_path = Path(settings.MEDIA_ROOT).joinpath(
        attachment_relative_path
    )
    
    # remove folder
    if attachment_full_path.exists():
        attachment_full_path.unlink()
    
    return attachment_full_path


"""add inner_id

Revision ID: e76c2ca5562e
Revises: 54fee6a7ecd8
Create Date: 2023-02-28 22:29:31.595257

"""
from alembic import op
from alembic import context
import sqlalchemy as sa
from sqlalchemy.sql import text
from labelu.alembic_labelu.alembic_labelu_tools import column_exist_in_table


# revision identifiers, used by Alembic.
revision = "e76c2ca5562e"
down_revision = "54fee6a7ecd8"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """
    add inner_id and last_sample_inner_id columns in the task and sample tables
    """
    with context.begin_transaction():
        if not column_exist_in_table(
            "task", "last_sample_inner_id"
        ):
            with op.batch_alter_table("task", recreate="always") as batch_op_task:
                batch_op_task.add_column(
                    sa.Column(
                        "last_sample_inner_id",
                        sa.Integer(),
                        nullable=True,
                        comment="The last inner id of sample in a task",
                        server_default=text("0"),
                    ),
                    insert_before="config",
                )
        if not column_exist_in_table("task_sample", "inner_id"):
            with op.batch_alter_table(
                "task_sample", recreate="always"
            ) as batch_op_task_sample:
                batch_op_task_sample.add_column(
                    sa.Column(
                        "inner_id",
                        sa.Integer(),
                        nullable=True,
                        comment="sample id in a task",
                    ),
                    insert_after="id",
                )

        op.execute(
            "UPDATE task SET last_sample_inner_id=(SELECT sample_number FROM ( SELECT task_id, COUNT(id) AS sample_number FROM task_sample GROUP BY task_id ) task_sample_number WHERE task.id=task_sample_number.task_id )"
        )
        op.execute(
            "UPDATE task_sample SET inner_id=(SELECT newid FROM ( SELECT id,  ROW_NUMBER() OVER (PARTITION BY task_id ORDER BY created_at) AS newid FROM task_sample) tmp_sample_inner_id WHERE task_sample.id=tmp_sample_inner_id.id)"
        )


def downgrade() -> None:
    """
    delete inner_id and last_sample_inner_id columns in the task and sample tables
    """
    op.drop_column("task_sample", "inner_id")
    op.drop_column("task", "last_sample_inner_id")



import os
from pathlib import Path

from loguru import logger
from pydantic import BaseSettings, Field

from labelu.internal.common.io import get_data_dir


class Settings(BaseSettings):
    SCHEME: str = "http"
    HOST: str = "localhost"
    PORT: str = "8000"
    API_V1_STR: str = "/api/v1"
    MEDIA_HOST: str = f"{SCHEME}://{HOST}:{PORT}"

    BASE_DATA_DIR: str = get_data_dir()
    MEDIA_ROOT: Path = Path(BASE_DATA_DIR).joinpath("media")
    UPLOAD_DIR: str = "upload"
    EXPORT_DIR: str = "export"
    os.makedirs(MEDIA_ROOT, exist_ok=True)
    logger.info("Database and media directory: {}", BASE_DATA_DIR)
    UPLOAD_FILE_MAX_SIZE: int = 200_000_000  # ~200MB
    THUMBNAIL_HEIGH_PIXEL: int = 120

    DATABASE_URL: str = Field(
        # default="mysql://labelu:labelupass@localhost/labeludb",
        default=f"sqlite:///{BASE_DATA_DIR}/labelu.sqlite",
        description="Database connection URL. Supports SQLite and MySQL."
    )
    # or using MySQL DATABASE_URL=mysql://labelu:labelupass@localhost/labeludb

    PASSWORD_SECRET_KEY: str = (
        "e5b7d00a59aaa2a5ea86a7c4d72f856b20bafa1b8d0e66124082ada81f6340bd"
    )

    TOKEN_GENERATE_ALGORITHM: str = "HS256"
    TOKEN_ACCESS_EXPIRE_MINUTES: int = 30
    TOKEN_TYPE: str = "Bearer"

    @property
    def need_migration_to_mysql(self) -> bool:
        sqlite_path = Path(self.BASE_DATA_DIR) / "labelu.sqlite"
        return (
            self.DATABASE_URL.startswith('mysql') and 
            sqlite_path.exists()
        )

    class Config:
        env_prefix = ""
        env_file_encoding = "utf-8"
        case_sensitive = True


settings = Settings()



from typing import Any
from loguru import logger
import uvicorn
from typer import Typer
from fastapi import FastAPI, Response
from fastapi.staticfiles import StaticFiles

from labelu.internal.adapter.routers import add_router
from labelu.internal.adapter.ws import add_ws_router
from labelu.internal.middleware import add_middleware
from labelu.internal.common.logger import init_logging
from labelu.internal.common.db import init_tables
from labelu.internal.common.config import settings
from labelu.internal.common.error_code import add_exception_handler
from labelu.alembic_labelu.run_migrate import run_db_migrations
from labelu.scripts.migrate_to_mysql import migrate_to_mysql

from .version import version as labelu_version

description = """
labelU backend.

## Users

You will be able to:

* **Signup**
* **Login**
* **Logout**.

## Tasks

You will be able to:

* **CRUD**

## Task attachment

You will be able to:

* **upload attachment**
* **download attachment**
* **delete attachment**

## Task sample

You will be able to:

* **list sample**
* **create sample**
* **get sample**
* **update sample**
* **export sample**
"""


tags_metadata = [
    {
        "name": "users",
        "description": "Operations with users.",
    },
    {
        "name": "tasks",
        "description": "Task management.",
    },
    {
        "name": "attachments",
        "description": "Task attachment management.",
    },
    {
        "name": "samples",
        "description": "Task sample management.",
    },
]

app = FastAPI(
    title="labelU",
    description=description,
    version=labelu_version,
    terms_of_service="",
    contact={
        "name": "labelu",
        "url": "http://labelu.example.com/contact/",
        "email": "labelu@example.com",
    },
    license_info={
        "name": "Apache 2.0",
        "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
    },
    openapi_tags=tags_metadata,
)

init_logging()
init_tables()
run_db_migrations()
add_exception_handler(app=app)
add_router(app=app)
add_ws_router(app=app)
add_middleware(app=app)

def startup():
    if settings.need_migration_to_mysql:
        logger.info("Migrating database to MySQL")
        migrate_to_mysql()

app.add_event_handler("startup", startup)

class NoCacheStaticFiles(StaticFiles):
    def __init__(self, *args: Any, **kwargs: Any):
        self.cachecontrol = "max-age=0, no-cache, no-store, must-revalidate"
        self.pragma = "no-cache"
        self.expires = "0"
        super().__init__(*args, **kwargs)

    def file_response(self, *args: Any, **kwargs: Any) -> Response:
        resp = super().file_response(*args, **kwargs)
        
        # No cache for html files
        if resp.media_type == "text/html":
            resp.headers.setdefault("Cache-Control", self.cachecontrol)
            resp.headers.setdefault("Pragma", self.pragma)
            resp.headers.setdefault("Expires", self.expires)
            
        return resp

app.mount("", NoCacheStaticFiles(packages=["labelu.internal"], html=True))

cli = Typer()

@cli.command('migrate_to_mysql')
def to_mysql():
    """Migrate database to MySQL"""
    migrate_to_mysql()

@cli.callback(invoke_without_command=True)
def main(
    host: str = "localhost", port: int = 8000, media_host: str = "http://localhost:8000"
):
    if port:
        settings.PORT = port
    if host:
        settings.HOST = host
    if media_host:
        settings.MEDIA_HOST = media_host
        
    uvicorn.run(app=app, host=settings.HOST, port=settings.PORT, ws="websockets")
        

if __name__ == "__main__":
    cli()


from datetime import datetime
from sqlalchemy.schema import Index
from sqlalchemy import Column, DateTime, ForeignKey, Integer
from labelu.internal.common.db import Base

class TaskSampleUpdater(Base):
    __tablename__ = "task_sample_updater"
    
    sample_id = Column(
        Integer, 
        ForeignKey("task_sample.id", ondelete="CASCADE"), 
        primary_key=True,
        index=True
    )
    user_id = Column(
        Integer, 
        ForeignKey("user.id", ondelete="CASCADE"), 
        primary_key=True,
        index=True
    )
    created_at = Column(
        DateTime(timezone=True),
        default=datetime.now,
        onupdate=datetime.now,
        nullable=False,
        comment="created time"
    )



from sqlalchemy.orm import Session
from fastapi.testclient import TestClient

from labelu.internal.common.config import settings
from labelu.internal.domain.models.task import Task
from labelu.internal.adapter.persistence import crud_user
from labelu.internal.adapter.persistence import crud_task


class TestClassTaskRouter:
    def test_create_task_successful(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }

        # run
        r = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )

        # check
        json = r.json()
        assert r.status_code == 201
        assert json["data"]["id"] > 0
        assert json["data"]["created_by"]["id"] > 0
        assert json["data"]["status"] == "DRAFT"

    def test_create_task_no_authentication(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }

        # run
        r = client.post(
            f"{settings.API_V1_STR}/tasks", headers=testuser_token_headers, json=data
        )

        # check
        assert r.status_code == 201

    def test_task_list(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        page = 0
        size = 10
        for i in range(15):
            crud_task.create(
                db=db,
                task=Task(
                    name="name",
                    description="description",
                    tips="tips",
                    created_by=current_user.id,
                    updated_by=current_user.id,
                ),
            )

        # run
        r = client.get(
            f"{settings.API_V1_STR}/tasks",
            headers=testuser_token_headers,
            params={"page": page, "size": size},
        )

        # check
        json = r.json()
        assert r.status_code == 200
        assert json["meta_data"]["total"] == 15
        assert len(json["data"]) == 10

    def test_task_get(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
            "config": "config",
            "media_type": "IMAGE",
        }
        task = client.post(
            f"{settings.API_V1_STR}/tasks",
            headers=testuser_token_headers,
            json=data,
        )

        # run
        task_id = task.json()["data"]["id"]
        r = client.get(
            f"{settings.API_V1_STR}/tasks/{task_id}",
            headers=testuser_token_headers,
        )

        # check
        json = r.json()["data"]
        assert r.status_code == 200
        assert json["name"] == "task name"
        assert json["created_by"]["id"] > 0

    def test_task_get_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        # run
        r = client.get(
            f"{settings.API_V1_STR}/tasks/0",
            headers=testuser_token_headers,
        )

        # check
        json = r.json()
        assert r.status_code == 404
        assert json["err_code"] == 50002

    def test_task_update(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        for_update_task_data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
            "config": "config",
            "media_type": "IMAGE",
        }
        for_update_task = client.post(
            f"{settings.API_V1_STR}/tasks",
            headers=testuser_token_headers,
            json=for_update_task_data,
        )
        for_updated_task_data = {
            "name": "new name",
            "description": "new description",
            "tips": "new tips",
            "config": "new config",
            "media_type": "IMAGE",
        }

        # run
        task_id = for_update_task.json()["data"]["id"]
        updated_task = client.patch(
            f"{settings.API_V1_STR}/tasks/{task_id}",
            headers=testuser_token_headers,
            json=for_updated_task_data,
        )

        # check
        json = updated_task.json()
        assert json["data"]["name"] == "new name"
        assert json["data"]["description"] == "new description"
        assert json["data"]["tips"] == "new tips"
        assert json["data"]["config"] == "new config"
        assert json["data"]["media_type"] == "IMAGE"

    def test_task_update_no_found_task(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        for_updated_task_data = {
            "name": "new name",
            "description": "new description",
            "tips": "new tips",
            "config": "new config",
            "media_type": "IMAGE",
        }

        # run
        updated_task = client.patch(
            f"{settings.API_V1_STR}/tasks/0",
            headers=testuser_token_headers,
            json=for_updated_task_data,
        )

        # check
        json = updated_task.json()
        assert updated_task.status_code == 404
        assert json["err_code"] == 50002

    def test_task_update_no_config(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        for_update_task_data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        for_update_task = client.post(
            f"{settings.API_V1_STR}/tasks",
            headers=testuser_token_headers,
            json=for_update_task_data,
        )
        for_updated_task_data = {
            "name": "new name",
            "description": "new description",
            "tips": "new tips",
            "config": "new config",
        }

        # run
        task_id = for_update_task.json()["data"]["id"]
        updated_task = client.patch(
            f"{settings.API_V1_STR}/tasks/{task_id}",
            headers=testuser_token_headers,
            json=for_updated_task_data,
        )

        # check
        json = updated_task.json()
        assert json["data"]["name"] == "new name"
        assert json["data"]["description"] == "new description"
        assert json["data"]["tips"] == "new tips"
        assert json["data"]["config"] == None
        assert json["data"]["media_type"] == None

    def test_task_add_annotation_config(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        for_update_task_data = {
            "name": "task name",
            "description": "task description",
            "tips": "task tips",
        }
        for_update_task = client.post(
            f"{settings.API_V1_STR}/tasks",
            headers=testuser_token_headers,
            json=for_update_task_data,
        )
        for_updated_task_data = {
            "name": "new name",
            "description": "new description",
            "tips": "new tips",
            "config": "new config",
            "media_type": "IMAGE",
        }

        # run
        task_id = for_update_task.json()["data"]["id"]
        updated_task = client.patch(
            f"{settings.API_V1_STR}/tasks/{task_id}",
            headers=testuser_token_headers,
            json=for_updated_task_data,
        )

        # check
        json = updated_task.json()
        assert json["data"]["name"] == "new name"
        assert json["data"]["description"] == "new description"
        assert json["data"]["tips"] == "new tips"
        assert json["data"]["config"] == "new config"
        assert json["data"]["media_type"] == "IMAGE"
        assert json["data"]["status"] == "CONFIGURED"

    def test_task_delete(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        current_user = crud_user.get_user_by_username(
            db=db, username="test@example.com"
        )
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=current_user.id,
                updated_by=current_user.id,
            ),
        )

        # run
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/{task.id}", headers=testuser_token_headers
        )

        # check
        json = r.json()
        assert r.status_code == 200
        assert json["data"]["ok"] == True

    def test_task_delete_not_found(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data

        # run
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/0", headers=testuser_token_headers
        )

        # check
        json = r.json()
        assert r.status_code == 404
        assert json["err_code"] == 50002

    def test_task_delete_no_permission(
        self, client: TestClient, testuser_token_headers: dict, db: Session
    ) -> None:

        # prepare data
        task = crud_task.create(
            db=db,
            task=Task(
                name="name",
                description="description",
                tips="tips",
                created_by=0,
                updated_by=0,
            ),
        )

        # run
        r = client.request("delete",
            f"{settings.API_V1_STR}/tasks/{task.id}", headers=testuser_token_headers
        )

        # check
        json = r.json()
        assert r.status_code == 403
        assert json["err_code"] == 30001



from loguru import logger
from starlette.requests import Request
from starlette.concurrency import iterate_in_threadpool
from starlette.middleware.base import BaseHTTPMiddleware


class TracingMiddleWare(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        scope = request.scope
        logger.info(f"{scope['path']} start")

        response = await call_next(request)

        # uncomments for logging response body
        # response_body = [chunk async for chunk in response.body_iterator]
        # response.body_iterator = iterate_in_threadpool(iter(response_body))
        # logger.info(f"response_body={(b''.join(response_body)).decode()}")

        logger.info(f"{scope['path']} end")
        return response



from datetime import datetime
from typing import List, Optional, Tuple

from sqlalchemy.orm import Session

from labelu.internal.domain.models.attachment import TaskAttachment

def list_by(
    db: Session,
    size: int,
    ids: List[int] | None = [],
    task_id: Optional[int] = None,
    owner_id: Optional[int] = None,
    after: Optional[int] = None,
    before: Optional[int] = None,
    page: Optional[int] = None,
    sorting: Optional[str] = None,
) -> Tuple[List[TaskAttachment], int]:
# query filter
    query_filter = [TaskAttachment.deleted_at == None, TaskAttachment.id.in_(ids)]
    if owner_id:
        query_filter.append(TaskAttachment.created_by == owner_id)
        
    if before:
        query_filter.append(TaskAttachment.id < before)
    if after:
        query_filter.append(TaskAttachment.id > after)
    if task_id:
        query_filter.append(TaskAttachment.task_id == task_id)
        
    query = db.query(TaskAttachment).filter(*query_filter)

    # default order by id, before need select last items
    if before:
        query = query.order_by(TaskAttachment.id.desc())
    else:
        query = query.order_by(TaskAttachment.id.asc())
        
    count = query.count()
    
    results = (
        query.offset(offset=page * size if page else 0)
        .limit(limit=size)
        .all()
    )
    
    if sorting:
        field, order = sorting.split(":")
        if order == "desc":
            results = sorted(results, key=lambda x: getattr(x, field), reverse=True)
        else:
            results = sorted(results, key=lambda x: getattr(x, field))
    
    if before:
        results.reverse()
        
    return results, count

def create(db: Session, attachment: TaskAttachment) -> TaskAttachment:
    db.add(attachment)
    db.flush()
    db.refresh(attachment)
    return attachment


def get(db: Session, attachment_id: int) -> TaskAttachment:
    return (
        db.query(TaskAttachment)
        .filter(TaskAttachment.id == attachment_id, TaskAttachment.deleted_at == None)
        .first()
    )


def get_by_ids(db: Session, attachment_ids: List[int]) -> List[TaskAttachment]:
    return (
        db.query(TaskAttachment)
        .filter(
            TaskAttachment.id.in_(attachment_ids), TaskAttachment.deleted_at == None
        )
        .all()
    )


def delete(db: Session, attachment_ids: List[int]) -> None:
    db.query(TaskAttachment).filter(TaskAttachment.id.in_(attachment_ids)).update(
        {TaskAttachment.deleted_at: datetime.now()}
    )



from fastapi import Depends, WebSocket, WebSocketDisconnect
from loguru import logger
from pydantic import BaseModel

from labelu.internal.clients.ws import sampleConnectionManager
from labelu.internal.common.websocket import ConnectionData, Message, MessageType
from labelu.internal.dependencies.user import verify_ws_token
    
class TaskSampleWsPayload(BaseModel):
    task_id: int
    user_id: int
    username: str
    sample_id: int
    
def get_task_sample_connection_payloads(conns: ConnectionData):
    if not conns:
        return Message(
            type=MessageType.PEERS,
            data=[]
        )
        
    return Message(
        type=MessageType.PEERS,
        data=[
            TaskSampleWsPayload(
                task_id=conn.data.task_id,
                user_id=conn.data.user_id,
                username=conn.data.username,
                sample_id=conn.data.sample_id
            ) for conn in conns
        ]
    )

    
async def task_ws_endpoint(websocket: WebSocket, task_id: int, sample_id: int, user=Depends(verify_ws_token)):
    if not user:
        await websocket.close(code=1008, reason="Unauthorized")
        return
    
    client_id = f"task_{task_id}"
    
    async def sync_peers():
        connections = sampleConnectionManager.active_connections
        current_connections = connections.get(client_id, {})
        sample_payload = get_task_sample_connection_payloads(current_connections.values())
        
        await sampleConnectionManager.send_message(client_id=client_id, message=sample_payload)
        
    
    connection = None
    
    async def cleanup():
        await sampleConnectionManager.disconnect(client_id, websocket)
        await sync_peers()
    
    try:
        connection = await sampleConnectionManager.connect(
            client_id,
            websocket,
            data=TaskSampleWsPayload(
                task_id=task_id,
                user_id=user.id,
                username=user.username,
                sample_id=sample_id,
            )
        )
        
        await sync_peers()
        
        while True:
            try:
                data = await websocket.receive_text()
                message = Message.parse_raw(data)
                
                if message.type == MessageType.PONG:
                    connection.update_heartbeat()
                
            except WebSocketDisconnect as e:
                logger.info(f"WebSocket disconnected for client {client_id}: Code={e.code}")
                break
            
    except WebSocketDisconnect as e:
        logger.info(f"WebSocket disconnected during connection setup for client {client_id}: Code={e.code}")
    except Exception as e:
        logger.error(f"WebSocket error for client {client_id}: {e}")
    finally:
        await cleanup()


from typing import Optional

from jose import jwt
from loguru import logger
from sqlalchemy.orm import Session
from pydantic import ValidationError
from fastapi import HTTPException, WebSocket, status, Depends, Request

from labelu.internal.domain.models.user import User
from labelu.internal.common import db
from labelu.internal.common.config import settings
from labelu.internal.common.security import AccessToken
from labelu.internal.common.error_code import ErrorCode
from labelu.internal.common.error_code import LabelUException
from labelu.internal.adapter.persistence import crud_user


class SpecialOAuth2PasswordBearer:
    def __call__(self, request: Request) -> Optional[str]:
        authorization: str = request.headers.get("Authorization")
        _, _, param = authorization.partition(" ")
        return param


reusable_oauth2 = SpecialOAuth2PasswordBearer()


def get_current_user(
    db: Session = Depends(db.get_db), token: str = Depends(reusable_oauth2)
) -> User:
    try:
        payload = jwt.decode(
            token,
            settings.PASSWORD_SECRET_KEY,
            algorithms=[settings.TOKEN_GENERATE_ALGORITHM],
            options={"verify_exp": False},
        )
        token_data = AccessToken(**payload)
    except (jwt.JWTError, ValidationError):
        raise LabelUException(
            code=ErrorCode.CODE_40003_CREDENTIAL_ERROR,
            status_code=status.HTTP_403_FORBIDDEN,
        )
    user = crud_user.get(db, id=token_data.id)
    if not user:
        raise LabelUException(code=ErrorCode.CODE_40002_USER_NOT_FOUND, status_code=401)
    return user


async def verify_ws_token(websocket: WebSocket, db: Session = Depends(db.get_db)):
    try:
        token = websocket.query_params.get('token')
        
        if not token:
            await websocket.close(code=4001, reason="Missing authentication token")
            raise HTTPException(status_code=401, detail="Missing authentication token")
            
        payload = jwt.decode(
            token,
            settings.PASSWORD_SECRET_KEY,
            algorithms=[settings.TOKEN_GENERATE_ALGORITHM],
            options={"verify_exp": False},
        )
        
        token_data = AccessToken(**payload)
        user = crud_user.get(db, id=token_data.id)
        
        if not user:
            await websocket.close(code=4002, reason="User not found")
            raise HTTPException(status_code=401, detail="User not found")
        
        return user
        
    except jwt.JWTError as e:
        logger.error(f"WebSocket authentication error: {str(e)}")
        await websocket.close(code=4003, reason="Invalid token")
        raise HTTPException(status_code=401, detail="Invalid token")